# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.nm.trainables.common.common_utils import *
from nemo.collections.nlp.nm.trainables.common.encoder_rnn import *
from nemo.collections.nlp.nm.trainables.common.huggingface import *
from nemo.collections.nlp.nm.trainables.common.sequence_classification_nm import *
from nemo.collections.nlp.nm.trainables.common.sequence_regression_nm import *
from nemo.collections.nlp.nm.trainables.common.token_classification_nm import *
from nemo.collections.nlp.nm.trainables.common.transformer import *
from nemo.utils import logging

try:
    from nemo.collections.nlp.nm.trainables.common.megatron.megatron_utils import *

except Exception as e:
    logging.error('Failed to import Megatron utils: `{}` ({})'.format(str(e), type(e)))
